from __future__ import print_function

import os
import getpass
import socket
import json
from baseline.reporting import EpochReportingHook
from mead.utils import read_config_file_or_json
from baseline.reporting import register_reporting
from xpctl.xpclient import Configuration
from xpctl.xpclient.api import XpctlApi
from xpctl.xpclient import ApiClient
from xpctl.xpclient.rest import ApiException
from xpctl.utils import to_swagger_experiment, store_model
from mead.utils import hash_config


@register_reporting(name='xpctl')
class XPCtlReporting(EpochReportingHook):
    def __init__(self, **kwargs):
        super(XPCtlReporting, self).__init__(**kwargs)
        # throw exception if the next three can't be read from kwargs
        if 'host' in kwargs:
            self.api_url = kwargs['host']
        elif 'cred' in kwargs:
            self.api_url = read_config_file_or_json(kwargs['cred'])['host']
        else:
            raise ValueError('must provide a url where xpctl server is running')
        self.exp_config = read_config_file_or_json(kwargs['config_file'])
        self.task = kwargs['task']
        
        self.label = kwargs.get('label', None)
        self.username = kwargs.get('user', getpass.getuser())
        self.hostname = kwargs.get('host', socket.gethostname())
        self.checkpoint_base = None
        self.checkpoint_store = kwargs.get('checkpoint_store', '/data/model-checkpoints')
        self.save_model = kwargs.get('save_model', False) # optionally save the model
        config = Configuration(host=self.api_url)
        api_client = ApiClient(config)
        self.api = XpctlApi(api_client)
        
        self.log = []
    
    def _step(self, metrics, tick, phase, tick_type, **kwargs):
        """Write intermediate results to a logging memory object that ll be pushed to the xpctl repo

        :param metrics: A map of metrics to scores
        :param tick: The time (resolution defined by `tick_type`)
        :param phase: The phase of training (`Train`, `Valid`, `Test`)
        :param tick_type: The resolution of tick (`STEP`, `EPOCH`)
        :return:
        """
        msg = {'tick_type': tick_type, 'tick': tick, 'phase': phase}
        for k, v in metrics.items():
            msg[k] = v
        self.log.append(msg)
    
    def done(self):
        """Write the log to the xpctl database"""
        try:
            result = self.api.put_result(
                self.task,
                to_swagger_experiment(self.task,
                                      self.exp_config,
                                      self.log,
                                      username=self.username,
                                      label=self.label,
                                      hostname=self.hostname,
                                      )
            )
            if result.response_type == 'failure':
                raise RuntimeError(result.message)
            else:
                print('result stored with experiment id', result.message)
            if self.save_model:
                eid = result.message
                backend = self.exp_config.get('backend', 'default')
                backends = {'default': 'tf', 'tensorflow': 'tf', 'pytorch': 'pyt'}
                self.checkpoint_base = self._search_checkpoint_base(self.task, backends[backend],
                                                                    self.exp_config.get('basedir'))
                if self.checkpoint_base is None:
                    raise RuntimeError('No checkpoint files found')
                result = store_model(checkpoint_base=self.checkpoint_base,
                                     config_sha1=hash_config(self.exp_config),
                                     checkpoint_store=self.checkpoint_store)
                if result is not None:
                    print('model stored at {}'.format(result))
                    update_result = self.api.update_property(self.task, eid, prop='checkpoint', value=result)
                    print(update_result.message)
                else:
                    raise RuntimeError('failed to store model at {}'.format(self.checkpoint_store))
        except ApiException as e:
            raise RuntimeError(json.loads(e.body)['detail'])
    
    @staticmethod
    def _search_checkpoint_base(task, backend, basedir=None):
        """Finds if the checkpoint exists as a zip file or a bunch of files."""
        if basedir is not None:
            zip = "{}-{}.zip".format(basedir, os.getpid())
            non_zip = "{}-{}".format(basedir, os.getpid())
        else:
            zip = "{}-model-{}-{}.zip".format(task, backend, os.getpid())
            non_zip = "{}-model-{}-{}".format(task, backend, os.getpid())
        if os.path.exists(zip):
            return zip
        elif os.path.exists(".graph".format(non_zip)):
            return non_zip
        return None


def create_reporting_hook(**kwargs):
    return XPCtlReporting(**kwargs)


# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from xpctl.xpserver.models.base_model_ import Model
from xpctl.xpserver import util


class TaskSummary(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, task=None, summary=None):  # noqa: E501
        """TaskSummary - a model defined in Swagger

        :param task: The task of this TaskSummary.  # noqa: E501
        :type task: str
        :param summary: The summary of this TaskSummary.  # noqa: E501
        :type summary: object
        """
        self.swagger_types = {
            'task': str,
            'summary': object
        }

        self.attribute_map = {
            'task': 'task',
            'summary': 'summary'
        }

        self._task = task
        self._summary = summary

    @classmethod
    def from_dict(cls, dikt):
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The TaskSummary of this TaskSummary.  # noqa: E501
        :rtype: TaskSummary
        """
        return util.deserialize_model(dikt, cls)

    @property
    def task(self):
        """Gets the task of this TaskSummary.


        :return: The task of this TaskSummary.
        :rtype: str
        """
        return self._task

    @task.setter
    def task(self, task):
        """Sets the task of this TaskSummary.


        :param task: The task of this TaskSummary.
        :type task: str
        """
        if task is None:
            raise ValueError("Invalid value for `task`, must not be `None`")  # noqa: E501

        self._task = task

    @property
    def summary(self):
        """Gets the summary of this TaskSummary.


        :return: The summary of this TaskSummary.
        :rtype: object
        """
        return self._summary

    @summary.setter
    def summary(self, summary):
        """Sets the summary of this TaskSummary.


        :param summary: The summary of this TaskSummary.
        :type summary: object
        """
        if summary is None:
            raise ValueError("Invalid value for `summary`, must not be `None`")  # noqa: E501

        self._summary = summary



# coding: utf-8

# flake8: noqa

"""
    xpctl

    This is a sample xpctl  server.  You can find out more about xpctl at [baseline](https://github.com/dpressel/baseline/blob/master/docs/xpctl.md).  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: apiteam@swagger.io
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from __future__ import absolute_import

# import apis into sdk package
from xpctl.xpclient.api.xpctl_api import XpctlApi

# import ApiClient
from xpctl.xpclient.api_client import ApiClient
from xpctl.xpclient.configuration import Configuration
# import models into sdk package
from xpctl.xpclient.models.aggregate_result import AggregateResult
from xpctl.xpclient.models.aggregate_result_values import AggregateResultValues
from xpctl.xpclient.models.experiment import Experiment
from xpctl.xpclient.models.experiment_aggregate import ExperimentAggregate
from xpctl.xpclient.models.response import Response
from xpctl.xpclient.models.result import Result
from xpctl.xpclient.models.task_summary import TaskSummary



# coding: utf-8

"""
    xpctl

    This is a sample xpctl  server.  You can find out more about xpctl at [baseline](https://github.com/dpressel/baseline/blob/master/docs/xpctl.md).  # noqa: E501

    OpenAPI spec version: 1.0.0
    Contact: apiteam@swagger.io
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from xpctl.xpclient.models.aggregate_result import AggregateResult  # noqa: F401,E501


class ExperimentAggregate(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'task': 'str',
        'eid': 'str',
        'sha1': 'str',
        'config': 'str',
        'dataset': 'str',
        'username': 'str',
        'hostname': 'str',
        'exp_date': 'date',
        'label': 'str',
        'version': 'str',
        'num_exps': 'int',
        'train_events': 'list[AggregateResult]',
        'valid_events': 'list[AggregateResult]',
        'test_events': 'list[AggregateResult]'
    }

    attribute_map = {
        'task': 'task',
        'eid': 'eid',
        'sha1': 'sha1',
        'config': 'config',
        'dataset': 'dataset',
        'username': 'username',
        'hostname': 'hostname',
        'exp_date': 'exp_date',
        'label': 'label',
        'version': 'version',
        'num_exps': 'num_exps',
        'train_events': 'train_events',
        'valid_events': 'valid_events',
        'test_events': 'test_events'
    }

    def __init__(self, task=None, eid=None, sha1=None, config=None, dataset=None, username=None, hostname=None, exp_date=None, label=None, version=None, num_exps=None, train_events=None, valid_events=None, test_events=None):  # noqa: E501
        """ExperimentAggregate - a model defined in Swagger"""  # noqa: E501

        self._task = None
        self._eid = None
        self._sha1 = None
        self._config = None
        self._dataset = None
        self._username = None
        self._hostname = None
        self._exp_date = None
        self._label = None
        self._version = None
        self._num_exps = None
        self._train_events = None
        self._valid_events = None
        self._test_events = None
        self.discriminator = None

        if task is not None:
            self.task = task
        if eid is not None:
            self.eid = eid
        if sha1 is not None:
            self.sha1 = sha1
        if config is not None:
            self.config = config
        if dataset is not None:
            self.dataset = dataset
        if username is not None:
            self.username = username
        if hostname is not None:
            self.hostname = hostname
        if exp_date is not None:
            self.exp_date = exp_date
        if label is not None:
            self.label = label
        if version is not None:
            self.version = version
        if num_exps is not None:
            self.num_exps = num_exps
        if train_events is not None:
            self.train_events = train_events
        if valid_events is not None:
            self.valid_events = valid_events
        if test_events is not None:
            self.test_events = test_events

    @property
    def task(self):
        """Gets the task of this ExperimentAggregate.  # noqa: E501


        :return: The task of this ExperimentAggregate.  # noqa: E501
        :rtype: str
        """
        return self._task

    @task.setter
    def task(self, task):
        """Sets the task of this ExperimentAggregate.


        :param task: The task of this ExperimentAggregate.  # noqa: E501
        :type: str
        """

        self._task = task

    @property
    def eid(self):
        """Gets the eid of this ExperimentAggregate.  # noqa: E501


        :return: The eid of this ExperimentAggregate.  # noqa: E501
        :rtype: str
        """
        return self._eid

    @eid.setter
    def eid(self, eid):
        """Sets the eid of this ExperimentAggregate.


        :param eid: The eid of this ExperimentAggregate.  # noqa: E501
        :type: str
        """

        self._eid = eid

    @property
    def sha1(self):
        """Gets the sha1 of this ExperimentAggregate.  # noqa: E501


        :return: The sha1 of this ExperimentAggregate.  # noqa: E501
        :rtype: str
        """
        return self._sha1

    @sha1.setter
    def sha1(self, sha1):
        """Sets the sha1 of this ExperimentAggregate.


        :param sha1: The sha1 of this ExperimentAggregate.  # noqa: E501
        :type: str
        """

        self._sha1 = sha1

    @property
    def config(self):
        """Gets the config of this ExperimentAggregate.  # noqa: E501


        :return: The config of this ExperimentAggregate.  # noqa: E501
        :rtype: str
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this ExperimentAggregate.


        :param config: The config of this ExperimentAggregate.  # noqa: E501
        :type: str
        """

        self._config = config

    @property
    def dataset(self):
        """Gets the dataset of this ExperimentAggregate.  # noqa: E501


        :return: The dataset of this ExperimentAggregate.  # noqa: E501
        :rtype: str
        """
        return self._dataset

    @dataset.setter
    def dataset(self, dataset):
        """Sets the dataset of this ExperimentAggregate.


        :param dataset: The dataset of this ExperimentAggregate.  # noqa: E501
        :type: str
        """

        self._dataset = dataset

    @property
    def username(self):
        """Gets the username of this ExperimentAggregate.  # noqa: E501


        :return: The username of this ExperimentAggregate.  # noqa: E501
        :rtype: str
        """
        return self._username

    @username.setter
    def username(self, username):
        """Sets the username of this ExperimentAggregate.


        :param username: The username of this ExperimentAggregate.  # noqa: E501
        :type: str
        """

        self._username = username

    @property
    def hostname(self):
        """Gets the hostname of this ExperimentAggregate.  # noqa: E501


        :return: The hostname of this ExperimentAggregate.  # noqa: E501
        :rtype: str
        """
        return self._hostname

    @hostname.setter
    def hostname(self, hostname):
        """Sets the hostname of this ExperimentAggregate.


        :param hostname: The hostname of this ExperimentAggregate.  # noqa: E501
        :type: str
        """

        self._hostname = hostname

    @property
    def exp_date(self):
        """Gets the exp_date of this ExperimentAggregate.  # noqa: E501


        :return: The exp_date of this ExperimentAggregate.  # noqa: E501
        :rtype: date
        """
        return self._exp_date

    @exp_date.setter
    def exp_date(self, exp_date):
        """Sets the exp_date of this ExperimentAggregate.


        :param exp_date: The exp_date of this ExperimentAggregate.  # noqa: E501
        :type: date
        """

        self._exp_date = exp_date

    @property
    def label(self):
        """Gets the label of this ExperimentAggregate.  # noqa: E501


        :return: The label of this ExperimentAggregate.  # noqa: E501
        :rtype: str
        """
        return self._label

    @label.setter
    def label(self, label):
        """Sets the label of this ExperimentAggregate.


        :param label: The label of this ExperimentAggregate.  # noqa: E501
        :type: str
        """

        self._label = label

    @property
    def version(self):
        """Gets the version of this ExperimentAggregate.  # noqa: E501


        :return: The version of this ExperimentAggregate.  # noqa: E501
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this ExperimentAggregate.


        :param version: The version of this ExperimentAggregate.  # noqa: E501
        :type: str
        """

        self._version = version

    @property
    def num_exps(self):
        """Gets the num_exps of this ExperimentAggregate.  # noqa: E501


        :return: The num_exps of this ExperimentAggregate.  # noqa: E501
        :rtype: int
        """
        return self._num_exps

    @num_exps.setter
    def num_exps(self, num_exps):
        """Sets the num_exps of this ExperimentAggregate.


        :param num_exps: The num_exps of this ExperimentAggregate.  # noqa: E501
        :type: int
        """

        self._num_exps = num_exps

    @property
    def train_events(self):
        """Gets the train_events of this ExperimentAggregate.  # noqa: E501


        :return: The train_events of this ExperimentAggregate.  # noqa: E501
        :rtype: list[AggregateResult]
        """
        return self._train_events

    @train_events.setter
    def train_events(self, train_events):
        """Sets the train_events of this ExperimentAggregate.


        :param train_events: The train_events of this ExperimentAggregate.  # noqa: E501
        :type: list[AggregateResult]
        """

        self._train_events = train_events

    @property
    def valid_events(self):
        """Gets the valid_events of this ExperimentAggregate.  # noqa: E501


        :return: The valid_events of this ExperimentAggregate.  # noqa: E501
        :rtype: list[AggregateResult]
        """
        return self._valid_events

    @valid_events.setter
    def valid_events(self, valid_events):
        """Sets the valid_events of this ExperimentAggregate.


        :param valid_events: The valid_events of this ExperimentAggregate.  # noqa: E501
        :type: list[AggregateResult]
        """

        self._valid_events = valid_events

    @property
    def test_events(self):
        """Gets the test_events of this ExperimentAggregate.  # noqa: E501


        :return: The test_events of this ExperimentAggregate.  # noqa: E501
        :rtype: list[AggregateResult]
        """
        return self._test_events

    @test_events.setter
    def test_events(self, test_events):
        """Sets the test_events of this ExperimentAggregate.


        :param test_events: The test_events of this ExperimentAggregate.  # noqa: E501
        :type: list[AggregateResult]
        """

        self._test_events = test_events

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ExperimentAggregate, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ExperimentAggregate):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



""" dump one xpctl database to local file system, restore back to another database"""
import argparse
import os
from xpctl.backends.core import ExperimentRepo
from baseline.utils import read_config_file


def main():
    parser = argparse.ArgumentParser(description='dump restore xpctl databases')
    parser.add_argument('--from_dbtype', help='backend type, `from` database', default='mongo')
    parser.add_argument('--from_cred', help='credential for backend, `from` database',
                        default=os.path.expanduser('~/xpctlcred.json'))
    parser.add_argument('--to_dbtype', help='backend type, `to` database', default='mongo')
    parser.add_argument('--to_cred', help='credential for backend, `to` database',
                        default=os.path.expanduser('~/xpctlcred-localhost.json'))
    args = parser.parse_args()
    
    d1 = read_config_file(args.from_cred)
    d1.update({'dbtype': args.from_dbtype})
    from_backend = ExperimentRepo().create_repo(**d1)
    dump_file = from_backend.dump()

    d2 = read_config_file(args.to_cred)
    d2.update({'dbtype': args.to_dbtype})
    to_backend = ExperimentRepo().create_repo(**d2)
    to_backend.restore(dump_file)
    

if __name__ == "__main__":
    main()



from __future__ import print_function
import os
import datetime
import socket
import json
import getpass
from baseline.utils import listify
from mead.utils import hash_config
from baseline.version import __version__
from xpctl.backends.backend import log2json, get_experiment_label, METRICS_SORT_ASCENDING, safe_get, \
    client_experiment_to_put_result_consumable, write_experiment, aggregate_results
from xpctl.backends.backend import BackendError, BackendSuccess, TaskDatasetSummary, TaskDatasetSummarySet, Experiment, Result, \
    ExperimentSet, EVENT_TYPES
from baseline.utils import unzip_files, read_config_file
import shutil


from sqlalchemy.ext.declarative import declarative_base
Base = declarative_base()
import sqlalchemy as sql
import sqlalchemy.orm as orm
from xpctl.backends.core import ExperimentRepo


class SqlResult(Base):
    __tablename__ = 'results'
    id = sql.Column(sql.Integer, primary_key=True)
    metric = sql.Column(sql.String, nullable=False)
    value = sql.Column(sql.Float, nullable=False)
    event_id = sql.Column(sql.Integer, sql.ForeignKey('events.id'))
    event = orm.relationship('SqlEvent', back_populates='results', cascade='all,delete')


class SqlEvent(Base):
    __tablename__ = 'events'
    id = sql.Column(sql.Integer, primary_key=True)
    phase = sql.Column(sql.String, nullable=False)
    tick_type = sql.Column(sql.String, nullable=False)
    tick = sql.Column(sql.Integer, nullable=False)
    experiment_id = sql.Column(sql.Integer, sql.ForeignKey('experiments.eid'), nullable=False)
    experiment = orm.relationship('SqlExperiment', back_populates='events', cascade='all,delete')
    results = orm.relationship('SqlResult', back_populates='event', cascade='all,delete')


class SqlExperiment(Base):
    __tablename__ = 'experiments'
    eid = sql.Column(sql.Integer, primary_key=True)
    label = sql.Column(sql.String)
    sha1 = sql.Column(sql.String)
    task = sql.Column(sql.String)
    dataset = sql.Column(sql.String)
    config = sql.Column(sql.String)
    hostname = sql.Column(sql.String)
    username = sql.Column(sql.String)
    version = sql.Column(sql.String)
    date = sql.Column(sql.DateTime)
    status = sql.Column(sql.String)
    last_modified = sql.Column(sql.DateTime)
    checkpoint = sql.Column(sql.String)
    events = orm.relationship('SqlEvent', back_populates='experiment', cascade='all,delete')


class SQLRepo(ExperimentRepo):

    def _connect(self, uri):
        self.engine = sql.create_engine(uri, echo=False, paramstyle='format', pool_size=100)
        Base.metadata.create_all(self.engine)
        self.Session = orm.sessionmaker(bind=self.engine)

    def __init__(self, **kwargs):
        super(SQLRepo, self).__init__()

        uri = kwargs.get('uri', None)
        if uri is None:
            dbtype = kwargs.get('dbtype', 'postgresql')
            host = kwargs.get('host', 'localhost')
            port = kwargs.get('port', None)
            if port is not None:
                host = '{}:{}'.format(host, port)
            username = kwargs.get('user', None)
            passwd = kwargs.get('passwd', None)

            user = username if passwd is None else '{}:{}'.format(username, passwd)
            dbname = kwargs.get('db', 'reporting_db')
            uri = '{}://{}@{}/{}'.format(dbtype, user, host, dbname)
        self._connect(uri)

    def put_result(self, task, exp):
        unpacked = client_experiment_to_put_result_consumable(exp)
        return self._put_result(task=task, config_obj=unpacked.config_obj, events_obj=unpacked.events_obj,
                                **unpacked.extra_args)

    def _put_result(self, task, config_obj, events_obj, **kwargs):
        session = self.Session()

        now = safe_get(kwargs, 'date', datetime.datetime.utcnow().isoformat())
        hostname = safe_get(kwargs, 'hostname', socket.gethostname())
        username = safe_get(kwargs, 'username', getpass.getuser())
        config_sha1 = safe_get(kwargs, 'sha1', hash_config(config_obj))
        label = safe_get(kwargs, 'label', get_experiment_label(config_obj, task, **kwargs))
        checkpoint = kwargs.get('checkpoint')
        version = safe_get(kwargs,  'version', __version__)
        dataset = safe_get(kwargs, 'dataset', config_obj.get('dataset'))
        date = safe_get(kwargs, 'exp_date', now)

        events = []
        for event in events_obj:
            tick = event['tick']
            phase = event['phase']
            event_obj = SqlEvent(phase=phase, tick_type=event['tick_type'], tick=tick, results=[])
            for key in event.keys():
                if key not in ['tick_type', 'tick', 'event_type', 'id', 'date', 'phase']:
                    metric = SqlResult(metric=key, value=event[key])
                    event_obj.results += [metric]
            events += [event_obj]
        experiment = SqlExperiment(
            label=label,
            checkpoint=checkpoint,
            sha1=config_sha1,
            task=task,
            dataset=dataset,
            config=json.dumps(config_obj),
            hostname=hostname,
            username=username,
            date=date,
            version=version,
            status='CREATED',
            last_modified=now,
            events=events
        )
        try:
            session.add(experiment)
            session.commit()
            return BackendSuccess(message=experiment.eid)
        except sql.exc.SQLAlchemyError as e:
            return BackendError(message=str(e))

    def get_model_location(self, task, eid):
        session = self.Session()
        exp = session.query(SqlExperiment).get(eid)
        return BackendSuccess(exp.checkpoint)

    def update_prop(self, task, eid, prop, value):
        try:
            session = self.Session()
            exp = session.query(SqlExperiment).filter(SqlExperiment.task == task).filter(SqlExperiment.eid == eid)
            if exp is None or exp.scalar() is None:
                return BackendError(message='no experiment with id [{}] for task [{}]'.format(eid, task))
            # there must be a better way of getting a column value through a column name
            prev_value = getattr(session.query(SqlExperiment).get(eid), prop)
            session.query(SqlExperiment).filter(SqlExperiment.eid == eid).update({prop: value})
            session.commit()
            changed_value = getattr(session.query(SqlExperiment).get(eid), prop)
            return BackendSuccess(message='for experiment [{}] property [{}] was changed from [{}] to [{}].'
                                  .format(eid, prop, prev_value, changed_value))
        except sql.exc.SQLAlchemyError as e:
            return BackendError(message=str(e))

    def remove_experiment(self, task, eid):
        try:
            session = self.Session()
            exp = session.query(SqlExperiment).filter(SqlExperiment.task == task).filter(SqlExperiment.eid == eid)
            if exp is None or exp.scalar() is None:
                return BackendError(message='no experiment with id [{}] for task [{}]'.format(eid, task))
            model_loc_response = self.get_model_location(task, eid)
            model_loc = model_loc_response.message
            if model_loc is not None and type(model_loc_response) is not BackendError and os.path.exists(model_loc):
                try:
                    os.remove(model_loc)
                except IOError:
                    return BackendError(message='model {} exists on host but could not be removed'.format(model_loc))
            exp = session.query(SqlExperiment).get(eid)
            session.delete(exp)
            session.commit()
            try:
                assert session.query(SqlExperiment).get(eid) is None
                return BackendSuccess("record [{}] deleted successfully from database [{}]".format(eid, task))
            except AssertionError:
                return BackendError('delete failed: could not delete experiment {} from {} database'.format(eid, task))
        except sql.exc.SQLAlchemyError as e:
            return BackendError(message=str(e))

    def get_experiment_details(self, task, eid, event_type, metric):
        metrics = [x for x in listify(metric) if x.strip()]
        if event_type is None or event_type == 'None':
            event_type = 'test_events'
        session = self.Session()
        exp = session.query(SqlExperiment).filter(SqlExperiment.task == task).filter(SqlExperiment.eid == eid)
        if exp is None or exp.scalar() is None:
            return BackendError(message='no experiment with id [{}] for task [{}]'.format(eid, task))
        return self.sql_result_to_data_experiment(exp.one(), event_type, metrics)

    def get_results(self, task, param_dict, reduction_dim, metric, sort, numexp_reduction_dim, event_type):
        session = self.Session()
        metrics = [x for x in listify(metric) if x.strip()]
        if event_type is None or event_type == 'None':
            event_type = 'test_events'
        reduction_dim = reduction_dim if reduction_dim is not None else 'sha1'
        if 'dataset' in param_dict.keys():
            param_dict['dataset'] = self.get_related_datasets(session, task, param_dict['dataset'])
        hits = session.query(SqlExperiment).filter(SqlExperiment.task == task)
        for prop, value in param_dict.items():
            if type(value) is list:
                hits = hits.filter(getattr(SqlExperiment, prop).in_(value))
            else:
                hits = hits.filter(getattr(SqlExperiment, prop) == value)
        if hits is None or not hits.first():
            return BackendError(message='no information available for {} in task database [{}]'
                                .format(param_dict, task))
        data_experiments = []
        for exp in hits:
            data_experiment = self.sql_result_to_data_experiment(exp, event_type, metrics)
            if type(data_experiment) is BackendError:
                return data_experiment
            else:
                data_experiments.append(data_experiment)
        experiment_aggregate_set = self.aggregate_sql_results(data_experiments, reduction_dim, event_type,
                                                              numexp_reduction_dim, param_dict)
        if sort is None or sort == 'None':
            return experiment_aggregate_set
        else:
            if event_type == 'test_events':
                if sort in METRICS_SORT_ASCENDING:
                    return experiment_aggregate_set.sort(sort, reverse=False)
                else:
                    return experiment_aggregate_set.sort(sort)
            else:
                return BackendError(message='experiments can only be sorted when event_type=test_events')

    def list_results(self, task, param_dict, user, metric, sort, event_type):
        session = self.Session()
        data_experiments = []
        if event_type is None or event_type == 'None':
            event_type = 'test_events'
        metrics = [x for x in listify(metric) if x.strip()]
        users = [x for x in listify(user) if x.strip()]
        if not param_dict:
            hits = session.query(SqlExperiment).filter(SqlExperiment.task == task)
        else:
            hits = session.query(SqlExperiment).filter(SqlExperiment.task == task)
            for prop, value in param_dict.items():
                hits = hits.filter(getattr(SqlExperiment, prop) == value)
        if users:
            hits = hits.filter(SqlExperiment.username.in_(users))
        if hits.first() is None:
            return BackendError('No results in {} database for {}'.format(task, param_dict))
        for exp in hits:
            data_experiment = self.sql_result_to_data_experiment(exp, event_type, metrics)
            if type(data_experiment) is BackendError:
                return data_experiment
            else:
                data_experiments.append(data_experiment)
        experiment_set = self.get_data_experiment_set(data_experiments)
        if sort is None or sort == 'None':
            return experiment_set
        else:
            if event_type == 'test_events':
                if sort in METRICS_SORT_ASCENDING:
                    return experiment_set.sort(sort, reverse=False)
                else:
                    return experiment_set.sort(sort)
            else:
                return BackendError(message='experiments can only be sorted when event_type=test_events')

    def find_experiments(self, task, prop, value):
        session = self.Session()
        sql_experiments = []
        hits = session.query(SqlExperiment).filter(getattr(SqlExperiment, prop) == value). \
            filter(SqlExperiment.task == task)
        if hits is None or not hits.first():
            return BackendError('No results in {} database for {} = {}'.format(task, prop, value))
        for exp in hits:
            sql_experiments.append(self.sql_result_to_data_experiment(exp, event_type='test_events', metrics_from_user=[]))
        return self.get_data_experiment_set(sql_experiments)

    def config2json(self, task, sha1):
        try:
            session = self.Session()
            config = session.query(SqlExperiment).filter(SqlExperiment.sha1 == sha1).one().config
            return json.loads(config)
        except sql.exc.SQLAlchemyError as e:
            return BackendError(message=str(e))

    def get_task_names(self):
        session = self.Session()
        return set([t[0] for t in session.query(SqlExperiment.task).distinct()])

    def task_summary(self, task):
        session = self.Session()
        datasets = [x[0] for x in session.query(SqlExperiment.dataset).distinct()]
        users = [x[0] for x in session.query(SqlExperiment.username).distinct()]
        store = []
        for dataset in datasets:
            user_num_exps = {}
            for user in users:
                result = len([x.username for x in session.query(SqlExperiment)
                             .filter(SqlExperiment.task == task)
                             .filter(SqlExperiment.dataset == dataset)
                             .filter(SqlExperiment.username == user).distinct()])
                if result != 0:
                    user_num_exps[user] = result
            store.append(TaskDatasetSummary(task=task, dataset=dataset, experiment_set=None, user_num_exps=user_num_exps))
        if not store:
            return BackendError('could not get summary for task: [{}]'.format(task))
        return TaskDatasetSummarySet(task=task, data=store).groupby()

    def summary(self):
        tasks = self.get_task_names()
        if "system.indexes" in tasks:
            tasks.remove("system.indexes")
        return [self.task_summary(task) for task in tasks]

    def dump(self, zipfile='xpctl-sqldump-{}'.format(datetime.datetime.now().isoformat()), task_eids={}):
        """ dump reporting log and config for later consumption. you"""
        tasks = self.get_task_names() if not task_eids.keys() else list(task_eids.keys())
        session = self.Session()

        base_dir = '/tmp/xpctldump'
        if os.path.exists(base_dir):
            shutil.rmtree(base_dir)

        os.makedirs(base_dir, exist_ok=True)
        for task in tasks:
            _dir = os.path.join(base_dir, task)
            os.makedirs(_dir)
            sql_exps = session.query(SqlExperiment).filter(SqlExperiment.task == task)
            for sql_exp in sql_exps:
                exp = self.sql_result_to_data_experiment(sql_exp, event_type=[], metrics_from_user=[])
                if type(exp) is BackendError:
                    print(exp.message)
                else:
                    write_experiment(exp, _dir)
        return shutil.make_archive(base_name=zipfile, format='zip', root_dir='/tmp', base_dir='xpctldump')

    def restore(self, dump):
        """ if dump is in zip format, will unzip it. expects the following dir structure in the unzipped file:
        <root>
         - <task>
           - <id>
             - <id>-reporting.log
             - <id>-config.yml
             - <id>-meta.yml (any meta info such as label, username etc.)
        """
        dump_dir = unzip_files(dump)
        for task in os.listdir(dump_dir):
            task_dir = os.path.join(dump_dir, task)
            for exp in os.listdir(task_dir):
                exp_dir = os.path.join(task_dir, exp)
                meta = [os.path.join(exp_dir, x) for x in os.listdir(exp_dir) if x.endswith('meta.yml')]
                reporting = [os.path.join(exp_dir, x) for x in os.listdir(exp_dir) if x.endswith('reporting.log')]
                config = [os.path.join(exp_dir, x) for x in os.listdir(exp_dir) if x.endswith('config.yml')]
                try:
                    assert len(config) == 1
                    assert len(reporting) == 1
                    assert len(meta) == 1
                    config = read_config_file(config[0])
                    meta = read_config_file(meta[0])
                    reporting = log2json(reporting[0])
                except AssertionError:
                    raise RuntimeError('There should be exactly one meta file, one config file and one reporting log '
                                       'in {}'.format(exp_dir))
                self._put_result(task, config_obj=config, events_obj=reporting, **meta)
        if dump_dir != dump:
            shutil.rmtree(dump_dir)

    @staticmethod
    def get_related_datasets(session, task, dataset):
        hits = session.query(SqlExperiment).filter(SqlExperiment.task == task)
        all_datasets = set([exp.dataset for exp in hits])
        if dataset in all_datasets:
            return [dataset]
        else:
            datasets = []
            for d in all_datasets:
                print(d)
                if d.startswith(dataset):
                    datasets.append(d)
            return datasets

    def aggregate_sql_results(self, data_experiments, reduction_dim, event_type, numexp_reduction_dim, prop_dict):
        experiment_set = self.get_data_experiment_set(data_experiments)
        return aggregate_results(experiment_set, reduction_dim, event_type, numexp_reduction_dim, prop_dict=prop_dict)

    def sql_result_to_data_experiment(self, exp, event_type, metrics_from_user):
        _exp = Experiment(
            task=exp.task,
            eid=exp.eid,
            username=exp.username,
            hostname=exp.hostname,
            config=exp.config,
            exp_date=exp.date,
            label=exp.label,
            dataset=exp.dataset,
            sha1=exp.sha1,
            version=exp.version,
            train_events=[],
            valid_events=[],
            test_events=[]
        )
        event_types = [event_type] if event_type else EVENT_TYPES
        for event_type in event_types:
            phase = self.event2phase(event_type)
            if type(phase) is BackendError:
                return phase
            phase_events = [event for event in exp.events if event.phase == phase]
            if len(phase_events) == 0:
                continue
            metrics = self.get_filtered_metrics(self.get_sql_metrics(phase_events[0]), set(metrics_from_user))
            if type(metrics) is BackendError:
                return metrics
            results = self.flatten([self.create_results(event, metrics) for event in phase_events])
            for r in results:
                _exp.add_result(r, event_type)
        return _exp

    @staticmethod
    def event2phase(event_type):
        if event_type == 'train_events':
            return 'Train'
        if event_type == 'valid_events':
            return 'Valid'
        if event_type == 'test_events':
            return 'Test'
        BackendError(message='Unknown event type {}'.format(event_type))

    @staticmethod
    def get_filtered_metrics(metrics_from_db, metrics_from_user):
        if not metrics_from_user:
            metrics = list(metrics_from_db)
        elif metrics_from_user - metrics_from_db:
            return BackendError(message='Metrics [{}] not found'.format(','.join(list(metrics_from_user - metrics_from_db))))
        else:
            metrics = list(metrics_from_user)
        return metrics

    @staticmethod
    def get_sql_metrics(event):
        return set([r.metric for r in event.results])

    @staticmethod
    def flatten(_list):
        return [item for sublist in _list for item in sublist]

    @staticmethod
    def create_results(event, metrics):
        results = []
        for r in event.results:
            if r.metric in metrics:
                results.append(
                    Result(metric=r.metric, value=r.value, tick_type=event.tick_type, tick=event.tick, phase=event.phase)
                )
        return results

    @staticmethod
    def get_data_experiment_set(data_experiments):
        return ExperimentSet(data_experiments)




# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from xpctl.xpserver.models.base_model_ import Model
from xpctl.xpserver import util


class Result(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, metric=None, value=None, tick_type=None, tick=None, phase=None):  # noqa: E501
        """Result - a model defined in Swagger

        :param metric: The metric of this Result.  # noqa: E501
        :type metric: str
        :param value: The value of this Result.  # noqa: E501
        :type value: float
        :param tick_type: The tick_type of this Result.  # noqa: E501
        :type tick_type: str
        :param tick: The tick of this Result.  # noqa: E501
        :type tick: int
        :param phase: The phase of this Result.  # noqa: E501
        :type phase: str
        """
        self.swagger_types = {
            'metric': str,
            'value': float,
            'tick_type': str,
            'tick': int,
            'phase': str
        }

        self.attribute_map = {
            'metric': 'metric',
            'value': 'value',
            'tick_type': 'tick_type',
            'tick': 'tick',
            'phase': 'phase'
        }

        self._metric = metric
        self._value = value
        self._tick_type = tick_type
        self._tick = tick
        self._phase = phase

    @classmethod
    def from_dict(cls, dikt):
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Result of this Result.  # noqa: E501
        :rtype: Result
        """
        return util.deserialize_model(dikt, cls)

    @property
    def metric(self):
        """Gets the metric of this Result.


        :return: The metric of this Result.
        :rtype: str
        """
        return self._metric

    @metric.setter
    def metric(self, metric):
        """Sets the metric of this Result.


        :param metric: The metric of this Result.
        :type metric: str
        """
        if metric is None:
            raise ValueError("Invalid value for `metric`, must not be `None`")  # noqa: E501

        self._metric = metric

    @property
    def value(self):
        """Gets the value of this Result.


        :return: The value of this Result.
        :rtype: float
        """
        return self._value

    @value.setter
    def value(self, value):
        """Sets the value of this Result.


        :param value: The value of this Result.
        :type value: float
        """
        if value is None:
            raise ValueError("Invalid value for `value`, must not be `None`")  # noqa: E501

        self._value = value

    @property
    def tick_type(self):
        """Gets the tick_type of this Result.


        :return: The tick_type of this Result.
        :rtype: str
        """
        return self._tick_type

    @tick_type.setter
    def tick_type(self, tick_type):
        """Sets the tick_type of this Result.


        :param tick_type: The tick_type of this Result.
        :type tick_type: str
        """
        if tick_type is None:
            raise ValueError("Invalid value for `tick_type`, must not be `None`")  # noqa: E501

        self._tick_type = tick_type

    @property
    def tick(self):
        """Gets the tick of this Result.


        :return: The tick of this Result.
        :rtype: int
        """
        return self._tick

    @tick.setter
    def tick(self, tick):
        """Sets the tick of this Result.


        :param tick: The tick of this Result.
        :type tick: int
        """
        if tick is None:
            raise ValueError("Invalid value for `tick`, must not be `None`")  # noqa: E501

        self._tick = tick

    @property
    def phase(self):
        """Gets the phase of this Result.


        :return: The phase of this Result.
        :rtype: str
        """
        return self._phase

    @phase.setter
    def phase(self, phase):
        """Sets the phase of this Result.


        :param phase: The phase of this Result.
        :type phase: str
        """
        if phase is None:
            raise ValueError("Invalid value for `phase`, must not be `None`")  # noqa: E501

        self._phase = phase



import connexion

from xpctl.xpserver.models.experiment import Experiment  # noqa: E501
import flask
from xpctl.backends.backend import serialize_dict, serialize_experiment, \
    serialize_experiment_aggregate, serialize_experiment_list, serialize_response, serialize_task_summary_list, \
    serialize_task_summary


def config2json(task, sha1):  # noqa: E501
    """get config for sha1

    config for sha1 # noqa: E501

    :param task: task
    :type task: str
    :param sha1: sha1
    :type sha1: str

    :rtype: object
    """
    backend = flask.globals.current_app.backend
    return serialize_dict(backend.config2json(task, sha1))


def experiment_details(task, eid, event_type=None, metric=None):  # noqa: E501
    """Find experiment by id

    Returns a single experiment # noqa: E501

    :param task: task name
    :type task: str
    :param eid: ID of experiment to return
    :type eid: str
    :param event_type: 
    :type event_type: str
    :param metric: 
    :type metric: List[str]

    :rtype: Experiment
    """
    backend = flask.globals.current_app.backend
    return serialize_experiment(backend.get_experiment_details(task, eid, event_type, metric))


def get_model_location(task, eid):  # noqa: E501
    """get model loc for experiment

    get model loc for experiment # noqa: E501

    :param task: task
    :type task: str
    :param eid: experiment id
    :type eid: str

    :rtype: Response
    """
    backend = flask.globals.current_app.backend
    return serialize_response(backend.get_model_location(task, eid))


def get_results_by_prop(task, eid=None, sha1=None, dataset=None, label=None, reduction_dim=None, metric=None, sort=None, numexp_reduction_dim=None, event_type=None):  # noqa: E501
    """Find results by property and value

    Find results by property and value # noqa: E501

    :param task: task name
    :type task: str
    :param eid:
    :type eid: str
    :param sha1:
    :type sha1: str
    :param dataset:
    :type dataset: str
    :param label:
    :type label: str
    :param reduction_dim: which dimension to reduce on, default&#x3D;sha1
    :type reduction_dim: str
    :param metric: metric
    :type metric: List[str]
    :param sort: metric to sort results on
    :type sort: str
    :param numexp_reduction_dim: max number of experiments in an aggregate group
    :type numexp_reduction_dim: int
    :param event_type: train/dev/test
    :type event_type: str

    :rtype: List[ExperimentAggregate]
    """
    backend = flask.globals.current_app.backend
    param_dict = {}
    if eid:
        param_dict['eid'] = eid
    if sha1:
        param_dict['sha1'] = sha1
    if dataset:
        param_dict['dataset'] = dataset
    if label:
        param_dict['label'] = label
    return serialize_experiment_aggregate(backend.get_results(task, param_dict, reduction_dim, metric, sort,
                                                              numexp_reduction_dim, event_type))


def list_experiments_by_prop(task, eid=None, sha1=None, dataset=None, label=None, user=None, metric=None, sort=None, event_type=None):  # noqa: E501
    """list all experiments for this property and value

    list all experiments for this property (sha1/ username) and value (1cd21df91770b4dbed64a683558b062e3dee61f0/ dpressel) # noqa: E501

    :param task: task name
    :type task: str
    :param eid:
    :type eid: str
    :param sha1:
    :type sha1: str
    :param dataset:
    :type dataset: str
    :param label:
    :type label: str
    :param user:
    :type user: List[str]
    :param metric: 
    :type metric: List[str]
    :param sort: 
    :type sort: str
    :param event_type: 
    :type event_type: str

    :rtype: List[Experiment]
    """
    backend = flask.globals.current_app.backend
    param_dict = {}
    if eid:
        param_dict['eid'] = eid
    if sha1:
        param_dict['sha1'] = sha1
    if dataset:
        param_dict['dataset'] = dataset
    if label:
        param_dict['label'] = label
    return serialize_experiment_list(backend.list_results(task, param_dict, user, metric, sort, event_type))


def put_result(task, experiment, user=None, label=None):  # noqa: E501
    """Add a new experiment in database

     # noqa: E501

    :param task: 
    :type task: str
    :param experiment: 
    :type experiment: dict | bytes
    :param user: 
    :type user: str
    :param label: 
    :type label: str

    :rtype: Response
    """
    if connexion.request.is_json:
        experiment = Experiment.from_dict(connexion.request.get_json())  # noqa: E501
    backend = flask.globals.current_app.backend
    return serialize_response(backend.put_result(task, experiment))


def remove_experiment(task, eid):  # noqa: E501
    """delete an experiment from the database

     # noqa: E501

    :param task: 
    :type task: str
    :param eid: 
    :type eid: str

    :rtype: Response
    """
    backend = flask.globals.current_app.backend
    return serialize_response(backend.remove_experiment(task, eid))


def summary():  # noqa: E501
    """get summary for task

    summary for task # noqa: E501


    :rtype: List[TaskSummary]
    """
    backend = flask.globals.current_app.backend
    return serialize_task_summary_list(backend.summary())


def task_summary(task):  # noqa: E501
    """get summary for task

    summary for task # noqa: E501

    :param task: task
    :type task: str

    :rtype: TaskSummary
    """
    backend = flask.globals.current_app.backend
    return serialize_task_summary(backend.task_summary(task))


def update_property(task, eid, prop, value):  # noqa: E501
    """update property for an experiment

     # noqa: E501

    :param task: 
    :type task: str
    :param eid: 
    :type eid: str
    :param prop: 
    :type prop: str
    :param value: 
    :type value: str

    :rtype: Response
    """
    backend = flask.globals.current_app.backend
    return serialize_response(backend.update_prop(task, eid, prop, value))



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from xpctl.xpserver.models.base_model_ import Model
from xpctl.xpserver.models.result import Result  # noqa: F401,E501
from xpctl.xpserver import util


class Experiment(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, task=None, eid=None, sha1=None, config=None, dataset=None, username=None, hostname=None, exp_date=None, label=None, version=None, train_events=None, valid_events=None, test_events=None):  # noqa: E501
        """Experiment - a model defined in Swagger

        :param task: The task of this Experiment.  # noqa: E501
        :type task: str
        :param eid: The eid of this Experiment.  # noqa: E501
        :type eid: str
        :param sha1: The sha1 of this Experiment.  # noqa: E501
        :type sha1: str
        :param config: The config of this Experiment.  # noqa: E501
        :type config: str
        :param dataset: The dataset of this Experiment.  # noqa: E501
        :type dataset: str
        :param username: The username of this Experiment.  # noqa: E501
        :type username: str
        :param hostname: The hostname of this Experiment.  # noqa: E501
        :type hostname: str
        :param exp_date: The exp_date of this Experiment.  # noqa: E501
        :type exp_date: str
        :param label: The label of this Experiment.  # noqa: E501
        :type label: str
        :param version: The version of this Experiment.  # noqa: E501
        :type version: str
        :param train_events: The train_events of this Experiment.  # noqa: E501
        :type train_events: List[Result]
        :param valid_events: The valid_events of this Experiment.  # noqa: E501
        :type valid_events: List[Result]
        :param test_events: The test_events of this Experiment.  # noqa: E501
        :type test_events: List[Result]
        """
        self.swagger_types = {
            'task': str,
            'eid': str,
            'sha1': str,
            'config': str,
            'dataset': str,
            'username': str,
            'hostname': str,
            'exp_date': str,
            'label': str,
            'version': str,
            'train_events': List[Result],
            'valid_events': List[Result],
            'test_events': List[Result]
        }

        self.attribute_map = {
            'task': 'task',
            'eid': 'eid',
            'sha1': 'sha1',
            'config': 'config',
            'dataset': 'dataset',
            'username': 'username',
            'hostname': 'hostname',
            'exp_date': 'exp_date',
            'label': 'label',
            'version': 'version',
            'train_events': 'train_events',
            'valid_events': 'valid_events',
            'test_events': 'test_events'
        }

        self._task = task
        self._eid = eid
        self._sha1 = sha1
        self._config = config
        self._dataset = dataset
        self._username = username
        self._hostname = hostname
        self._exp_date = exp_date
        self._label = label
        self._version = version
        self._train_events = train_events
        self._valid_events = valid_events
        self._test_events = test_events

    @classmethod
    def from_dict(cls, dikt):
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Experiment of this Experiment.  # noqa: E501
        :rtype: Experiment
        """
        return util.deserialize_model(dikt, cls)

    @property
    def task(self):
        """Gets the task of this Experiment.


        :return: The task of this Experiment.
        :rtype: str
        """
        return self._task

    @task.setter
    def task(self, task):
        """Sets the task of this Experiment.


        :param task: The task of this Experiment.
        :type task: str
        """

        self._task = task

    @property
    def eid(self):
        """Gets the eid of this Experiment.


        :return: The eid of this Experiment.
        :rtype: str
        """
        return self._eid

    @eid.setter
    def eid(self, eid):
        """Sets the eid of this Experiment.


        :param eid: The eid of this Experiment.
        :type eid: str
        """

        self._eid = eid

    @property
    def sha1(self):
        """Gets the sha1 of this Experiment.


        :return: The sha1 of this Experiment.
        :rtype: str
        """
        return self._sha1

    @sha1.setter
    def sha1(self, sha1):
        """Sets the sha1 of this Experiment.


        :param sha1: The sha1 of this Experiment.
        :type sha1: str
        """

        self._sha1 = sha1

    @property
    def config(self):
        """Gets the config of this Experiment.


        :return: The config of this Experiment.
        :rtype: str
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this Experiment.


        :param config: The config of this Experiment.
        :type config: str
        """

        self._config = config

    @property
    def dataset(self):
        """Gets the dataset of this Experiment.


        :return: The dataset of this Experiment.
        :rtype: str
        """
        return self._dataset

    @dataset.setter
    def dataset(self, dataset):
        """Sets the dataset of this Experiment.


        :param dataset: The dataset of this Experiment.
        :type dataset: str
        """

        self._dataset = dataset

    @property
    def username(self):
        """Gets the username of this Experiment.


        :return: The username of this Experiment.
        :rtype: str
        """
        return self._username

    @username.setter
    def username(self, username):
        """Sets the username of this Experiment.


        :param username: The username of this Experiment.
        :type username: str
        """

        self._username = username

    @property
    def hostname(self):
        """Gets the hostname of this Experiment.


        :return: The hostname of this Experiment.
        :rtype: str
        """
        return self._hostname

    @hostname.setter
    def hostname(self, hostname):
        """Sets the hostname of this Experiment.


        :param hostname: The hostname of this Experiment.
        :type hostname: str
        """

        self._hostname = hostname

    @property
    def exp_date(self):
        """Gets the exp_date of this Experiment.


        :return: The exp_date of this Experiment.
        :rtype: str
        """
        return self._exp_date

    @exp_date.setter
    def exp_date(self, exp_date):
        """Sets the exp_date of this Experiment.


        :param exp_date: The exp_date of this Experiment.
        :type exp_date: str
        """

        self._exp_date = exp_date

    @property
    def label(self):
        """Gets the label of this Experiment.


        :return: The label of this Experiment.
        :rtype: str
        """
        return self._label

    @label.setter
    def label(self, label):
        """Sets the label of this Experiment.


        :param label: The label of this Experiment.
        :type label: str
        """

        self._label = label

    @property
    def version(self):
        """Gets the version of this Experiment.


        :return: The version of this Experiment.
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this Experiment.


        :param version: The version of this Experiment.
        :type version: str
        """

        self._version = version

    @property
    def train_events(self):
        """Gets the train_events of this Experiment.


        :return: The train_events of this Experiment.
        :rtype: List[Result]
        """
        return self._train_events

    @train_events.setter
    def train_events(self, train_events):
        """Sets the train_events of this Experiment.


        :param train_events: The train_events of this Experiment.
        :type train_events: List[Result]
        """

        self._train_events = train_events

    @property
    def valid_events(self):
        """Gets the valid_events of this Experiment.


        :return: The valid_events of this Experiment.
        :rtype: List[Result]
        """
        return self._valid_events

    @valid_events.setter
    def valid_events(self, valid_events):
        """Sets the valid_events of this Experiment.


        :param valid_events: The valid_events of this Experiment.
        :type valid_events: List[Result]
        """

        self._valid_events = valid_events

    @property
    def test_events(self):
        """Gets the test_events of this Experiment.


        :return: The test_events of this Experiment.
        :rtype: List[Result]
        """
        return self._test_events

    @test_events.setter
    def test_events(self, test_events):
        """Sets the test_events of this Experiment.


        :param test_events: The test_events of this Experiment.
        :type test_events: List[Result]
        """

        self._test_events = test_events



import os
import shutil
import pymongo
import datetime
import socket
import getpass
from baseline.utils import listify, unzip_files, read_config_file
from mead.utils import hash_config
from xpctl.backends.core import ExperimentRepo
from xpctl.backends.backend import TaskDatasetSummary, TaskDatasetSummarySet, BackendSuccess, BackendError, Experiment, ExperimentSet, Result, \
    EVENT_TYPES, log2json, get_experiment_label, METRICS_SORT_ASCENDING, safe_get, \
    client_experiment_to_put_result_consumable, aggregate_results, write_experiment
from bson.objectid import ObjectId
from baseline.version import __version__
import logging


class MongoRepo(ExperimentRepo):

    def __init__(self, dbhost, dbport, user, passwd):
        super(MongoRepo, self).__init__()
        self.logger = logging.getLogger('xpctl-mongo')
        self.dbhost = dbhost
        if user and passwd:
            uri = "mongodb://{}:{}@{}:{}/test".format(user, passwd, dbhost, dbport)
            client = pymongo.MongoClient(uri)
        else:
            client = pymongo.MongoClient(dbhost, dbport)
        if client is None:
            s = "cannot connect to mongo at host: [{}], port [{}], username: [{}], password: [{}]".format(dbhost,
                                                                                                          dbport,
                                                                                                          user,
                                                                                                          passwd)
            raise Exception(s)
        try:
            dbnames = client.list_database_names()
        except pymongo.errors.ServerSelectionTimeoutError:
            raise Exception("cannot get database from mongo at host: {}, port {}, connection timed out".format(dbhost,
                                                                                                               dbport))

        if "reporting_db" not in dbnames:
            self.logger.warning("Warning: database reporting_db does not exist, do not query before inserting results")
        self.db = client.reporting_db

    def put_result(self, task, exp):
        unpacked = client_experiment_to_put_result_consumable(exp)
        return self._put_result(task=task, config_obj=unpacked.config_obj, events_obj=unpacked.events_obj,
                                **unpacked.extra_args)

    def _put_result(self, task, config_obj, events_obj, **kwargs):
        now = safe_get(kwargs, 'date', datetime.datetime.utcnow().isoformat())
        hostname = safe_get(kwargs, 'hostname', socket.gethostname())
        username = safe_get(kwargs, 'username', getpass.getuser())
        config_sha1 = safe_get(kwargs, 'sha1', hash_config(config_obj))
        label = safe_get(kwargs, 'label', get_experiment_label(config_obj, task, **kwargs))
        checkpoint = kwargs.get('checkpoint')
        version = safe_get(kwargs, 'version', __version__)
        dataset = safe_get(kwargs, 'dataset', config_obj.get('dataset'))
        date = safe_get(kwargs, 'exp_date', now)
        train_events = list(filter(lambda x: x['phase'] == 'Train', events_obj))
        valid_events = list(filter(lambda x: x['phase'] == 'Valid', events_obj))
        test_events = list(filter(lambda x: x['phase'] == 'Test', events_obj))

        post = {
            "dataset": dataset,
            "config": config_obj,
            "train_events": train_events,
            "valid_events": valid_events,
            "test_events": test_events,
            "username": username,
            "hostname": hostname,
            "date": date,
            "label": label,
            "sha1": config_sha1,
            "version": version,
            "checkpoint": checkpoint
        }
        if 'eid' in kwargs:
            post.update({'_id': ObjectId(kwargs['eid'])})

        try:
            coll = self.db[task]
            result = coll.insert_one(post)
            return BackendSuccess(message=str(result.inserted_id))
        except pymongo.errors.PyMongoError as e:
            return BackendError(message='experiment could not be inserted: {}'.format(e.message))

    def get_model_location(self, task, eid):
        coll = self.db[task]
        query = {'_id': ObjectId(eid)}
        projection = {'checkpoint': 1}
        results = [x.get('checkpoint') for x in list(coll.find(query, projection))]
        results = [x for x in results if x is not None]
        if not results:
            return BackendError(message='no model location for experiment id [{}] in [{}] database'.format(eid, task))
        return BackendSuccess(results[0])

    def update_prop(self, task, eid, prop, value):
        try:
            coll = self.db[task]
            r = coll.find_one({'_id': ObjectId(eid)}, {prop: 1})
            if r is None:
                return BackendError(message='property {} for experiment {} not found in {} database'.format(prop, eid, task))
            prev_value = r[prop]
            coll.update({'_id': ObjectId(eid)}, {'$set': {prop: value}}, upsert=False)
            changed_value = coll.find_one({'_id': ObjectId(eid)}, {prop: 1})[prop]
            return BackendSuccess(message='for experiment [{}] property [{}] was changed from [{}] to [{}]'
                                  .format(eid, prop, prev_value, changed_value))
        except pymongo.errors.PyMongoError as e:
            return BackendError(message='property update failed: {}'.format(e.message))

    def remove_experiment(self, task, eid):
        try:
            coll = self.db[task]
            prev = coll.find_one({'_id': ObjectId(eid)})
            if prev is None:
                return BackendError(message='delete operation failed: experiment [{}] not found in [{}] database'.format(eid, task))
            model_loc_response = self.get_model_location(task, eid)
            model_loc = model_loc_response.message
            if model_loc is not None and type(model_loc_response) is not BackendError and os.path.exists(model_loc):
                try:
                    os.remove(model_loc)
                except IOError:
                    return BackendError(message='model {} exists on host but could not be removed'.format(model_loc))
            coll.remove({'_id': ObjectId(eid)})
            try:
                assert coll.find_one({'_id': ObjectId(eid)}) is None
                return BackendSuccess("record [{}] deleted successfully from database [{}]".format(eid, task))
            except AssertionError:
                return BackendError('delete failed: could not delete experiment {} from {} database'.format(eid, task))
        except pymongo.errors.PyMongoError as e:
            return BackendError(message='experiment could not be removed: {}'.format(e.message))

    def get_experiment_details(self, task, eid, event_type, metric):
        metrics = [x for x in listify(metric) if x.strip()]
        if event_type is None or event_type == 'None':
            event_type = 'test_events'
        coll = self.db[task]
        query = {'_id': ObjectId(eid)}
        all_results = list(coll.find(query))
        if not all_results:
            return BackendError(message='no experiment with id [{}] for task [{}]'.format(eid, task))
        experiments = mongo_to_experiment_set(task, all_results, event_type=event_type, metrics=metrics)
        if type(experiments) == BackendError:
            return BackendError(experiments.message)
        return experiments[0]

    def get_results(self, task, param_dict, reduction_dim, metric, sort, numexp_reduction_dim, event_type):
        metrics = [x for x in listify(metric)]
        if event_type is None or event_type == 'None':
            event_type = 'test_events'
        reduction_dim = reduction_dim if reduction_dim is not None else 'sha1'
        coll = self.db[task]
        if 'dataset' in param_dict.keys():
            value = self.get_related_datasets(task, param_dict['dataset'])
            param_dict['dataset'] = value
        query = self._update_query({}, **param_dict)
        all_results = list(coll.find(query))
        if not all_results:
            return BackendError(message='no information available for {} in task database [{}]'
                                .format(param_dict, task))
        resultset = mongo_to_experiment_set(task, all_results, event_type=event_type, metrics=metrics)
        if type(resultset) is BackendError:
            return resultset
        experiment_aggregate_set = aggregate_results(resultset, reduction_dim, event_type, numexp_reduction_dim,
                                                     param_dict)
        if sort is None or sort == 'None':
            return experiment_aggregate_set
        else:
            if event_type == 'test_events':
                if sort in METRICS_SORT_ASCENDING:
                    return experiment_aggregate_set.sort(sort, reverse=False)
                else:
                    return experiment_aggregate_set.sort(sort)
            else:
                return BackendError(message='experiments can only be sorted when event_type=test_events')

    @staticmethod
    def _update_query(q, **kwargs):
        query = q
        if not kwargs:
            return query
        else:
            if "id" in kwargs and kwargs["id"]:
                if type(kwargs["id"]) is list:
                    query.update({"_id": {"$in": [ObjectId(x) for x in kwargs["id"]]}})
                else:
                    query.update({"_id": ObjectId(kwargs["id"])})
                kwargs.pop("id")
            if "eid" in kwargs and kwargs["eid"]:
                if type(kwargs["eid"]) is list:
                    query.update({"_id": {"$in": [ObjectId(x) for x in kwargs["eid"]]}})
                else:
                    query.update({"_id": ObjectId(kwargs["eid"])})
                kwargs.pop("eid")
            if "dataset" in kwargs:
                if type(kwargs["dataset"]) is list:
                    query.update({"$or": [
                        {"config.dataset": {"$in": kwargs["dataset"]}},
                        {"dataset": {"$in": kwargs["dataset"]}}
                    ]})
                else:
                    query.update({"$or": [{"config.dataset": kwargs["dataset"]}, {"dataset": kwargs["dataset"]}]})
                kwargs.pop("dataset")
            for key, value in kwargs.items():
                if type(value) == list:
                    query.update({key: {"$in": value}})
                else:
                    query.update({key: value})
            return query

    @staticmethod
    def _update_projection(event_type):
        projection = {"_id": 1, "sha1": 1, "label": 1, "username": 1, "config.dataset": 1, "date": 1}
        projection.update({event_type: 1})
        return projection

    def list_results(self, task, param_dict, user, metric, sort, event_type):
        if event_type is None or event_type == 'None':
            event_type = 'test_events'
        metrics = [x for x in listify(metric) if x.strip()]
        users = [x for x in listify(user) if x.strip()]
        if users:
            param_dict.update({'username': users})
        coll = self.db[task]
        query = self._update_query({}, **param_dict)
        all_results = list(coll.find(query))
        if not all_results:
            return BackendError(message='no information available for {} in task database [{}]'
                                .format(param_dict, task))
        experiments = mongo_to_experiment_set(task, all_results, event_type=event_type, metrics=metrics)
        if type(experiments) == BackendError:
            return experiments
        if sort is None or sort == 'None':
            return experiments
        else:
            if event_type == 'test_events':
                if sort in METRICS_SORT_ASCENDING:
                    return experiments.sort(sort, reverse=False)
                else:
                    return experiments.sort(sort)
            else:
                return BackendError(message='experiments can only be sorted when event_type=test_events')

    def find_experiments(self, task, prop, value):
        d = {prop: value}
        coll = self.db[task]
        query = self._update_query({}, **d)
        all_results = list(coll.find(query))
        if not all_results:
            return BackendError(message='no information available for [{}]: [{}] in task database [{}]'
                                .format(prop, value, task))
        return mongo_to_experiment_set(task, all_results, event_type='test_events', metrics=[])

    def config2json(self, task, sha1):
        coll = self.db[task]
        j = coll.find_one({"sha1": sha1}, {"config": 1})
        if not j:
            return BackendError('no config [{}] in [{}] database'.format(sha1, task))
        else:
            return j["config"]

    def get_related_datasets(self, task, dataset):
        """
        for now, lets just get the datasets from all tasks and filter by 'project:name:features'
        We should have a dataset collection in the future.
        :param dataset:
        :return:
        """
        coll = self.db[task]
        projection = {"dataset": 1, "config.dataset": 1, "_id": 0}
        results = list(coll.find({}, projection))
        all_datasets = []
        for result in results:
            all_datasets.append(result.get('dataset'))
            all_datasets.append(result.get('config', {}).get('dataset'))
        all_datasets = set([x for x in all_datasets if x is not None])
        if dataset in all_datasets:
            return [dataset]
        else:
            datasets = []
            for d in all_datasets:
                if d.startswith(dataset):
                    datasets.append(d)
            return datasets

    def get_task_names(self):
        return self.db.collection_names()

    def task_summary(self, task):
        event_type = 'test_events'
        coll = self.db[task]
        q = {}
        p = {'config.dataset': 1}
        datasets = list(set([x['config']['dataset'] for x in list(coll.find(q, p))]))
        store = []
        for dataset in datasets:
            q = self._update_query({}, dataset=dataset)
            p = self._update_projection(event_type)
            results = list(coll.find(q, p))
            experiment_set = mongo_to_experiment_set(task, results, event_type, metrics=[])
            if type(experiment_set) == BackendError:
                self.logger.error('Error getting summary for task [{}], dataset [{}], stacktrace [{}]'
                                  .format(task, dataset, experiment_set.message))
                continue
            store.append(TaskDatasetSummary(task=task, dataset=dataset, experiment_set=experiment_set))
        if not store:
            return BackendError('could not get summary for task: [{}]'.format(task))
        return TaskDatasetSummarySet(task=task, data=store).groupby()

    def summary(self):
        tasks = self.get_task_names()
        if "system.indexes" in tasks:
            tasks.remove("system.indexes")
        return [self.task_summary(task) for task in tasks]

    def dump(self, zipfile='xpctldump-{}'.format(datetime.datetime.now().isoformat()), task_eids={}):
        """ dump reporting log and config for later consumption"""
        tasks = self.get_task_names() if not task_eids.keys() else list(task_eids.keys())
        if "system.indexes" in tasks:
            tasks.remove("system.indexes")

        base_dir = '/tmp/xpctldump'
        if os.path.exists(base_dir):
            shutil.rmtree(base_dir)

        os.makedirs(base_dir, exist_ok=True)

        for task in tasks:
            coll = self.db[task]
            query = self._update_query({}, id=listify(task_eids.get(task, [])))
            all_results = list(coll.find(query))
            experiments = mongo_to_experiment_set(task, all_results, event_type=[], metrics=[]).data
            _dir = os.path.join(base_dir, task)
            os.makedirs(_dir)
            for exp in experiments:
                write_experiment(exp, _dir)
        return shutil.make_archive(base_name=zipfile, format='zip', root_dir='/tmp', base_dir='xpctldump')

    def restore(self, dump):
        """ if dump is in zip format, will unzip it. expects the following dir structure in the unzipped file:
        <root>
         - <task>
           - <id>-reporting.log
           - <id>.yml
        """
        dump_dir = unzip_files(dump)
        for task in os.listdir(dump_dir):
            task_dir = os.path.join(dump_dir, task)
            for exp in os.listdir(task_dir):
                exp_dir = os.path.join(task_dir, exp)
                meta = [os.path.join(exp_dir, x) for x in os.listdir(exp_dir) if x.endswith('meta.yml')]
                reporting = [os.path.join(exp_dir, x) for x in os.listdir(exp_dir) if x.endswith('reporting.log')]
                config = [os.path.join(exp_dir, x) for x in os.listdir(exp_dir) if x.endswith('config.yml')]
                try:
                    assert len(config) == 1
                    assert len(reporting) == 1
                    assert len(meta) == 1
                    config = read_config_file(config[0])
                    meta = read_config_file(meta[0])
                    reporting = log2json(reporting[0])
                except AssertionError:
                    raise RuntimeError('There should be exactly one meta file, one config file and one reporting log '
                                       'in {}'.format(exp_dir))
                self._put_result(task, config_obj=config, events_obj=reporting, **meta)
        if dump_dir != dump:
            shutil.rmtree(dump_dir)


class MongoResult(object):
    """ a result data point"""
    def __init__(self, metric, value, task, eid, username, hostname, label, config, dataset, date, sha1, event_type,
                 tick_type, tick, phase, version):
        super(MongoResult, self).__init__()
        self.metric = metric
        self.value = value
        self.task = task
        self.eid = eid
        self.username = username
        self.hostname = hostname
        self.label = label
        self.dataset = dataset
        self.date = date
        self.sha1 = sha1
        self.event_type = event_type
        self.config = config
        self.tick_type = tick_type
        self.tick = tick
        self.phase = phase
        self.version = version

    def get_prop(self, field):
        return self.__dict__[field]


class MongoResultSet(object):
    """ a list of result objects"""
    def __init__(self, data):
        super(MongoResultSet, self).__init__()
        self.data = data if data else []
        self.length = len(self.data)

    def add_data(self, data_point):
        """
        add a result data point
        :param data_point:
        :return:
        """
        self.data.append(data_point)
        self.length += 1

    # TODO: add property annotations
    def __getitem__(self, i):
        return self.data[i]

    def __iter__(self):
        for i in range(self.length):
            yield self.data[i]

    def __len__(self):
        return self.length

    def groupby(self, key):
        """ group the data points by key"""
        data_groups = {}
        for datum in self.data:
            field = datum.get_prop(key)
            if field not in data_groups:
                data_groups[field] = [datum]
            else:
                data_groups[field].append(datum)
        return data_groups

    def experiments(self):
        grouped_results = self.groupby('eid')
        experiments = []
        for eid, resultset in grouped_results.items():
            first_result = resultset[0]
            task = first_result.task
            eid = str(first_result.eid)
            username = first_result.username
            hostname = first_result.hostname
            label = first_result.label
            dataset = first_result.dataset
            date = first_result.date
            sha1 = first_result.sha1
            config = first_result.config
            version = first_result.version
            exp = Experiment(
                task=task,
                eid=eid,
                sha1=sha1,
                config=config,
                dataset=dataset,
                username=username,
                hostname=hostname,
                exp_date=date,
                label=label,
                version=version,
                train_events=[],
                valid_events=[],
                test_events=[])
            for _result in resultset:
                r = Result(metric=_result.metric, value=_result.value, tick_type=_result.tick_type, tick=_result.tick,
                           phase=_result.phase)
                exp.add_result(r, _result.event_type)
            experiments.append(exp)
        return ExperimentSet(experiments)


def get_metrics_mongo(xs):
    keys = []
    for x in xs:
        keys += x.keys()
    keys = set(keys)
    if 'tick_type' in keys:
        keys.remove("tick_type")
    if 'tick' in keys:
        keys.remove("tick")
    if 'phase' in keys:
        keys.remove("phase")
    return keys


def mongo_to_experiment_set(task, all_results, event_type, metrics):
    data = []
    event_types = [event_type] if event_type else EVENT_TYPES
    metrics_from_user = set([x for x in metrics if x.strip()])
    for result in all_results:  # different experiments
        task = task
        _id = result['_id']
        username = result.get('username', 'root')
        hostname = result.get('hostname', 'localhost')
        label = result.get('label', 'default_label')
        dataset = result.get('dataset', result['config'].get('dataset'))
        date = result['date']
        sha1 = result['sha1']
        config = result['config']
        version = result.get('version', '0.5.0')  # backward compatibility
        for event_type in event_types:
            if not result.get(event_type, []):
                continue
            metrics_from_db = get_metrics_mongo(result[event_type])
            if not metrics_from_user:
                metrics = list(metrics_from_db)
            elif metrics_from_user - metrics_from_db:
                return BackendError(message='Metrics [{}] not found for experiment [{}] in [{}] database'.format(','.join(
                    list(metrics_from_user - metrics_from_db)), _id, task))
            else:
                metrics = list(metrics_from_user)
            # for train_events we can have different metrics than test_events
            for record in result[event_type]:  # train_event epoch 0,
                for metric in metrics:
                    data.append(MongoResult(
                        metric=metric,
                        value=record[metric],
                        task=task,
                        eid=str(_id),
                        username=username,
                        hostname=hostname,
                        label=label,
                        config=config,
                        dataset=dataset,
                        date=date,
                        sha1=sha1,
                        event_type=event_type,
                        tick_type=record['tick_type'],
                        tick=record['tick'],
                        phase=record['phase'],
                        version=version
                    ))
    if not data:
        return BackendError(message='No results from the query')
    rs = MongoResultSet(data=data)
    return rs.experiments()



import os
import shutil
from baseline.utils import unzip_model, read_config_file, write_json, str_file
import json
import yaml
from xpctl.xpclient import Configuration, ApiClient
from xpctl.xpclient.api import XpctlApi
from xpctl.xpclient.models import Result, Experiment


def read_logs(file_name):
    logs = []
    with open(file_name) as f:
        for line in f:
            logs.append(json.loads(line))
    return logs


def convert_to_result(event):
    results = []
    non_metrics = ['tick_type', 'tick', 'phase']
    metrics = event.keys() - non_metrics
    for metric in metrics:
        results.append(Result(
            metric=metric,
            value=event[metric],
            tick_type=event['tick_type'],
            tick=event['tick'],
            phase=event['phase']
        )
        )
    return results


def flatten(_list):
    return [item for sublist in _list for item in sublist]


def to_swagger_experiment(task, config, log, **kwargs):
    if type(log) is not str:  # this is a log object and not a file
        events_obj = log
    else:
        events_obj = read_logs(log)
    train_events = flatten(
        [convert_to_result(event) for event in list(filter(lambda x: x['phase'] == 'Train', events_obj))]
    )
    valid_events = flatten(
        [convert_to_result(event) for event in list(filter(lambda x: x['phase'] == 'Valid', events_obj))]
    )
    test_events = flatten(
        [convert_to_result(event) for event in list(filter(lambda x: x['phase'] == 'Test', events_obj))]
    )
    if type(config) is not str:  # this is a config object and not a file
        config = json.dumps(config)
    else:
        config = json.dumps(read_config_file(config))
    d = kwargs
    d.update({'task': task,
              'config': config,
              'train_events': train_events,
              'valid_events': valid_events,
              'test_events': test_events
              })
    
    return Experiment(**d)


def store_model(checkpoint_base, config_sha1, checkpoint_store, print_fn=print):
    checkpoint_base = unzip_model(checkpoint_base)
    mdir, mbase = os.path.split(checkpoint_base)
    mdir = mdir if mdir else "."
    if not os.path.exists(mdir):
        print_fn("no directory found for the model location: [{}], aborting command".format(mdir))
        return None
    
    mfiles = ["{}/{}".format(mdir, x) for x in os.listdir(mdir) if x.startswith(mbase + "-") or
              x.startswith(mbase + ".")]
    if not mfiles:
        print_fn("no model files found with base [{}] at location [{}], aborting command".format(mbase, mdir))
        return None
    model_loc_base = "{}/{}".format(checkpoint_store, config_sha1)
    if not os.path.exists(model_loc_base):
        os.makedirs(model_loc_base)
    dirs = [int(x[:-4]) for x in os.listdir(model_loc_base) if x.endswith(".zip") and x[:-4].isdigit()]
    # we expect dirs in numbers.
    new_dir = "1" if not dirs else str(max(dirs) + 1)
    model_loc = "{}/{}".format(model_loc_base, new_dir)
    os.makedirs(model_loc)
    for mfile in mfiles:
        shutil.copy(mfile, model_loc)
        print_fn("writing model file: [{}] to store: [{}]".format(mfile, model_loc))
    print_fn("zipping model files")
    shutil.make_archive(base_name=model_loc,
                        format='zip',
                        root_dir=model_loc_base,
                        base_dir=new_dir)
    shutil.rmtree(model_loc)
    print_fn("model files zipped and written")
    return model_loc + ".zip"


def write_config_file(content, filepath):
    """Write a config file. This method optionally supports YAML, if the dependency was already installed.  O.W. JSON plz
    :param content: config object
    :param filepath: (``str``) A path to a file which should be a JSON file, or YAML if pyyaml is installed
    :return:
    """
    if filepath.endswith('.yml') or filepath.endswith('.yaml'):
        return write_yaml(content, filepath)
    return write_json(content, filepath)


@str_file(filepath="w")
def write_yaml(content, filepath):
    yaml.dump(content, filepath, default_flow_style=False)


def xpctl_client(host):
    config = Configuration(host)
    api_client = ApiClient(config)
    return XpctlApi(api_client)



import json
import pytest
from xpctl.xpclient import Configuration
from xpctl.xpclient.api import XpctlApi
from xpctl.xpclient import ApiClient
from xpctl.xpclient.models import Experiment, Result
from xpctl.xpclient.rest import ApiException
from mead.utils import hash_config
from collections import namedtuple
import datetime
import random
import string
from functools import wraps

TASK = 'test_xpctl'
HOST = 'localhost:5310/v2'
API = None


def clean_db(func):
    @wraps(func)
    def clean(*args, **kwargs):
        try:
            eids = [x.eid for x in API.list_experiments_by_prop(TASK)]
            for eid in eids:
                response = API.remove_experiment(TASK, eid)
                if response.response_type == 'error':
                    raise RuntimeError('can not clean db')
        except ApiException:
            # there's probably nothing in db
            return func(*args, **kwargs)
        return func(*args, **kwargs)
    return clean


@pytest.fixture(scope='session', autouse=True)
def setup():
    config = Configuration(host=HOST)
    api_client = ApiClient(config)
    global API
    API = XpctlApi(api_client)


def _put_one_exp(exp):
    result = API.put_result(TASK, exp)
    if result.response_type == 'error':
        raise RuntimeError(result.message)
    else:
        return result.message


@clean_db
def test_put_result_blank(setup):
    """ you can not insert an experiment with no results"""
    exp = Experiment(
        config='{"test": "test"}',
        train_events=[],
        valid_events=[],
        test_events=[]
    )
    try:
        _ = _put_one_exp(exp)
    except ApiException:
        return True


@clean_db
def test_put_result_one(setup):
    """ test a proper insertion"""
    exp = Experiment(
        config='{"test": "test"}',
        train_events=[],
        valid_events=[],
        test_events=[Result(metric='t', tick_type='t', phase='Test', tick=0, value=0.1)]
    )
    eid = _put_one_exp(exp)
    result = API.list_experiments_by_prop(TASK, eid=eid)
    assert len(result) == 1


@clean_db
def test_experiment_details(setup):
    date = datetime.datetime.utcnow().isoformat()
    label = 'test_label'
    dataset = 'test_dataset'
    config = '{"test":"test"}'
    username = 'user'
    hostname = 'host'
    exp = Experiment(
        dataset=dataset,
        label=label,
        exp_date=date,
        config=config,
        username=username,
        hostname=hostname,
        train_events=[],
        valid_events=[],
        test_events=[Result(metric='t', tick_type='t', phase='Test', tick=0, value=0.1)]
    )
    eid = _put_one_exp(exp)
    try:
        result = API.list_experiments_by_prop(TASK, eid=eid)
    except ApiException:
        print(eid)
        return False
    exp = result[0]
    assert exp.dataset == 'test_dataset'
    assert exp.label == label
    assert (exp.exp_date == date) or (exp.exp_date[:-1] == date)  # sql inserts Z at the end
    assert exp.sha1 == hash_config(json.loads('{"test": "test"}'))
    assert exp.username == username
    assert exp.hostname == hostname


def _generate_random_string(num_chars=20):
    return ''.join(random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits)
                   for _ in range(num_chars))


def _find_by_prop_unique(prop, expected_value):
    """
    since the property has an unique value, we should see just one experiment with that property
    :param prop: 'dataset', 'username' etc
    :param expected_value: a value generated from _generate_random_string
    :return:
    """
    try:
        if prop == 'label':
            result = API.list_experiments_by_prop(TASK, label=expected_value)
        elif prop == 'dataset':
            result = API.list_experiments_by_prop(TASK, dataset=expected_value)
        elif prop == 'username':
            result = API.list_experiments_by_prop(TASK, user=[expected_value])
        else:
            return False
        return len(result) == 1
    except ApiException:
        return False


@clean_db
def test_update_property(setup):
    date = datetime.datetime.utcnow().isoformat()
    label = _generate_random_string()
    dataset = _generate_random_string()
    config = '{"test":"test"}'
    username = _generate_random_string()
    exp = Experiment(
        dataset='d',
        label='l',
        exp_date=date,
        config=config,
        username='u',
        hostname='h',
        train_events=[],
        valid_events=[],
        test_events=[Result(metric='t', tick_type='t', phase='Test', tick=0, value=0.1)]
    )
    eid = _put_one_exp(exp)
    API.update_property(TASK, eid, 'dataset', dataset)
    API.update_property(TASK, eid, 'label', label)
    API.update_property(TASK, eid, 'username', username)
    assert _find_by_prop_unique(prop='label', expected_value=label)
    assert _find_by_prop_unique(prop='dataset', expected_value=dataset)
    assert _find_by_prop_unique(prop='username', expected_value=username)


def _test_list_experiments_by_prop(prop, value, experiments):
    if prop == 'dataset':
        results = API.list_experiments_by_prop(TASK, dataset=value)
    elif prop == 'label':
        results = API.list_experiments_by_prop(TASK, label=value)
    elif prop == 'username':
        results = API.list_experiments_by_prop(TASK, user=[value])
    else:
        return False
    expected = [exp.eid for exp in experiments if getattr(exp, prop) == value]
    results = [r.eid for r in results]
    return set(expected) == set(results)


@clean_db
def test_list_experiments_by_prop(setup):
    """
    list_experiments_by_prop finds some experiments by a (prop, value) pair. also allows some filtering:
    :param setup:
    :return:
    """
    configs = ['{"c1":"c1"}', '{"c2":"c2"}']
    datasets = ['d1', 'd2']
    labels = ['l1', 'l2']
    users = ['u1', 'u2', 'u3']
    metrics = ['f1', 'acc', 'random']
    test_events = [Result(metric=metric, value=0.5, tick_type='EPOCH', tick=0, phase='Test') for metric in metrics]
    experiments = []
    exp_detail = namedtuple('exp_detail', ['eid', 'sha1', 'dataset', 'label', 'username'])
    for config in configs:
        for dataset in datasets:
            for label in labels:
                for username in users:
                    result = _put_one_exp(
                        Experiment(
                            config=config,
                            dataset=dataset,
                            label=label,
                            username=username,
                            train_events=[],
                            valid_events=[],
                            test_events=test_events
                        )
                    )
                    experiments.append(exp_detail(eid=result, sha1=hash_config(json.loads(config)), dataset=dataset,
                                                  label=label, username=username))

    # find by a property and group by different reduction dims
    for prop, values in {'dataset': datasets, 'label': labels, 'username': users}.items():
        for value in values:
            assert _test_list_experiments_by_prop(prop=prop, value=value, experiments=experiments)

    # test the `users` filter work
    prop = 'dataset'
    value = 'd1'
    users = [['u1', 'u2'], ['u1']]
    for user in users:
        results = API.list_experiments_by_prop(TASK, dataset=value, user=user)
        result_eids = [x.eid for x in results]
        expected_eids = [x.eid for x in experiments if x.dataset == 'd1' and x.username in user]
        assert set(result_eids) == set(expected_eids)
    # test sort works
    config = configs[0]
    metrics = ['f1', 'acc']
    exp_value = namedtuple('exp_value', ['eid', 'value'])
    experiments = []
    dataset = _generate_random_string()  # generate a unique dataset
    for value in [0.5, 0.7, 0.8]:
        test_events = [Result(metric=metric, value=value, tick_type='EPOCH', tick=0, phase='Test') for metric in metrics]
        result = _put_one_exp(
            Experiment(
                config=config,
                train_events=[],
                valid_events=[],
                test_events=test_events,
                dataset=dataset
            )
        )
        experiments.append(exp_value(eid=result, value=value))
    results = API.list_experiments_by_prop(TASK, dataset=dataset, sort='f1')  # get results only from that
    # unique dataset, sort by f1
    _max = experiments[-1]
    assert results[0].eid == _max.eid


def _test_reduction_dim_dataset(dataset_value, reduction_dim, experiments):
    """
    The results from API should return proper groups
    :param prop:
    :param value:
    :param reduction_dim:
    :param experiments:
    :return:
    """
    results = API.get_results_by_prop(TASK, dataset=dataset_value, reduction_dim=reduction_dim)
    expected = {}
    for item in [getattr(exp, reduction_dim) for exp in experiments if getattr(exp, 'dataset') == dataset_value]:
        if item not in expected:
            expected[item] = 1
        else:
            expected[item] += 1
    results = {getattr(r, reduction_dim): r.num_exps for r in results}
    return expected == results


def _test_reduction_dim_label(label_value, reduction_dim, experiments):
    """
    The results from API should return proper groups
    :param prop:
    :param value:
    :param reduction_dim:
    :param experiments:
    :return:
    """
    results = API.get_results_by_prop(TASK, label=label_value, reduction_dim=reduction_dim)
    expected = {}
    for item in [getattr(exp, reduction_dim) for exp in experiments if getattr(exp, 'label') == label_value]:
        if item not in expected:
            expected[item] = 1
        else:
            expected[item] += 1
    results = {getattr(r, reduction_dim): r.num_exps for r in results}
    return expected == results


@clean_db
def test_get_results_by_prop(setup):
    """
    get_results_by_prop first finds some experiments by a (prop, value) pair and then groups them by a reduction_dim.
    typically we use prop='dataset' and reduction_dim='sha1', but the API supports more general calls
    :param setup:
    :return:
    """
    configs = ['{"c1":"c1"}', '{"c2":"c2"}']
    datasets = ['d1', 'd1', 'd2']
    labels = ['l1', 'l2', 'l2']
    metrics = ['f1', 'acc', 'random']
    test_events = [Result(metric=metric, value=0.5, tick_type='EPOCH', tick=0, phase='Test') for metric in metrics]
    experiments = []
    exp_detail = namedtuple('exp_detail', ['eid', 'sha1', 'dataset', 'label'])
    for config in configs:
        for dataset in datasets:
            for label in labels:
                result = _put_one_exp(
                    Experiment(
                        config=config,
                        dataset=dataset,
                        label=label,
                        train_events=[],
                        valid_events=[],
                        test_events=test_events
                    )
                )
                experiments.append(exp_detail(eid=result, sha1=hash_config(json.loads(config)), dataset=dataset,
                                              label=label))

    # find by a property and group by different reduction dims
    for prop, values in {'dataset': datasets, 'label': labels}.items():
        for value in values:
            for reduction_dim in ['sha1', 'eid', 'label']:
                if prop == 'dataset':
                    assert _test_reduction_dim_dataset(dataset_value=value, reduction_dim=reduction_dim, experiments=experiments)
                else:
                    assert _test_reduction_dim_label(label_value=value, reduction_dim=reduction_dim, experiments=experiments)
